package org.owasp.csrfguard;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Enumeration;
import java.security.NoSuchAlgorithmException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.htmlparser.Node;
import org.htmlparser.Parser;
import org.htmlparser.nodes.TagNode;
import org.htmlparser.tags.FormTag;
import org.htmlparser.tags.InputTag;
import org.htmlparser.tags.LinkTag;
import org.htmlparser.tags.ScriptTag;
import org.htmlparser.util.NodeIterator;
import org.htmlparser.util.NodeList;
import org.htmlparser.util.ParserException;

import org.owasp.csrfguard.http.MutableHttpResponse;
import org.owasp.csrfguard.util.CryptoUtil;

public final class CSRFGuard implements Filter
{
	private final static String PARAM_ERROR_PAGE = "error-page";
	
	private final static String PARAM_TOKEN_NAME = "token-name";
	
	private final static String PARAM_TOKEN_LENGTH = "token-length";
	
	private final static String PARAM_PRNG_ALGORITHM = "prng-algorithm";
	
	private final static String PARAM_DEBUG = "debug";
	
	private ServletContext context = null;
	
	private String errorPage = null;
	
	private String tokenName = null;
	
	private int tokenLength = -1;
	
	private String prngAlgorithm = null;
	
	private boolean debug = false;
	
	public void init(FilterConfig config)
	{
		context = config.getServletContext();
		errorPage = getErrorPage(config);
		tokenName = getTokenName(config);
		tokenLength = getTokenLength(config);
		prngAlgorithm = getPRNGAlgorithm(config);
		debug = getDebug(config);
		
		if(isDebugMode())
		{
			context.log("[CSRFGuard] init(): errorPage=" + errorPage + ",tokenName=" + tokenName + ",tokenLength=" + tokenLength + ",prngAlgorithm=" + prngAlgorithm + ",debug=" + debug);
		}
	}
	
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain)
	{
		if(request instanceof HttpServletRequest && response instanceof HttpServletResponse)
		{
			HttpServletRequest hRequest = (HttpServletRequest)request;
			MutableHttpResponse mResponse = new MutableHttpResponse((HttpServletResponse)response);
			String uri = hRequest.getRequestURI();
			
			if(isDebugMode())
			{
				context.log("[CSRFGuard] processing " + hRequest.getMethod() + " request to resource " + uri);
			}
			
			try
			{
				doRequestEnforce(hRequest);
				chain.doFilter(hRequest, mResponse);
				
				doResponseEnforce(hRequest, mResponse);
				doResponse(hRequest, mResponse);
			}
			catch(CSRFException e)
			{
				doError(hRequest, mResponse);
				
				context.log(e.getMessage());
			}
			catch (Exception e)
			{
				context.log(e.getMessage(), e);
			}
		}
	}
	
	public void destroy()
	{
		
	}
	
    private void doError(HttpServletRequest request, MutableHttpResponse response)
    {
    	HttpSession session = request.getSession(false);
    	HttpServletResponse hResponse = response.getOriginalResponse();
    	
    	if(session != null) { session.invalidate(); }
    	
    	try
    	{
    		hResponse.sendRedirect(errorPage);
    	}
    	catch(Exception e)
    	{
    		context.log(e.getMessage(), e);
    	}
    }
	
    private void doResponse(HttpServletRequest request, MutableHttpResponse response) throws IOException, ServletException
    {
    	
    	HttpServletResponse hResponse = response.getOriginalResponse();
        OutputStream out = hResponse.getOutputStream();
        byte[] result = response.getContent();
        
        if(isDebugMode())
    	{
    		context.log("[CSRFGuard] doResponse: writing content of type " + response.getContentType() + " with length " + result.length);
    	}
        
        response.setContentLength(result.length);
        out.write(result);
        out.close();
    }
    
    private void doRequestEnforce(HttpServletRequest request) throws CSRFException, NoSuchAlgorithmException, UnsupportedEncodingException
    {
    	HttpSession session = request.getSession(false);
    	
    	if(session != null)
    	{
    		String oToken = getTokenFromSession(session);
        	String nToken = getTokenFromRequest(request);
        	
        	if(isDebugMode())
        	{
        		context.log("[CSRFGuard] doRequestEnforce: oToken=" + oToken);
        		context.log("[CSRFGuard] doRequestEnforce: nToken="+ nToken);
        	}
        	
        	if(oToken == null)
        	{
        		setToken(session);
        	}
        	else if(!oToken.equals(nToken))
        	{
        		String ip = request.getRemoteAddr();
        		String method = request.getMethod();
        		String uri = CryptoUtil.URLEncode(request.getRequestURI());
        		String referer = CryptoUtil.URLEncode(request.getHeader("referer"));
        		String parameters = CryptoUtil.URLEncode(getRequestParameters(request));
        		throw new CSRFException("[CSRFGuard] caught CSRF attack (IP: " + ip + " Method: " + method + " URI: " + uri + " Referer: " + referer + " Parameters: " + parameters + ")");
        	}
    	}
    	else if(isDebugMode())
    	{
    		context.log("[CSRFGuard] doRequestEnforce: unable to verify the token as the session does not exist");
    	}
    }
    
    private String getTokenFromSession(HttpSession session)
    {
    	return (String)session.getAttribute(tokenName);
    }
    
    private String getTokenFromRequest(HttpServletRequest request)
    {
    	/**
    	 * The htmlparser appears to add newline characters in the token when it exceeds a certain length.
    	 * Frequently happens in links. This method makes sure the token value only has characters that
    	 * BASE64 can produce.
    	 */
    	String s = request.getParameter(tokenName);
    	StringBuffer sb = new StringBuffer();
    	
    	for(int i=0; i<s.length(); i++)
    	{
    		char c = s.charAt(i);
    		
    		if(c>= 32 && c <= 126)
    		{
    			sb.append(c);
    		}
    	}
    	
    	return sb.toString();
    }
    
    private String getRequestParameters(HttpServletRequest request)
    {
    	StringBuffer sb = new StringBuffer();
    	Enumeration<?> e = request.getParameterNames();
    	
    	//if(!e.hasMoreElements()) { sb.append("null"); }
    	
    	while(e.hasMoreElements())
    	{
    		String name = (String)e.nextElement();
    		String[] values = request.getParameterValues(name);
    		
    		for(int i=0; i<values.length; i++)
    		{
    			sb.append(name);
    			sb.append("=");
    			sb.append(values[i]);
    			
    			if((i+1) < values.length)
    			{
    				sb.append(",");
    			}
    		}
    		
    		if(e.hasMoreElements())
    		{
    			sb.append(",");
    		}
    	}
    	
    	return sb.toString();
    }
    
	private void setToken(HttpSession session) throws NoSuchAlgorithmException
	{
		String token = CryptoUtil.generateRandomToken(prngAlgorithm, tokenLength).replace("+", "_");
		session.setAttribute(tokenName, token);
	}
	
    private void doResponseEnforce(HttpServletRequest request, MutableHttpResponse response) throws IOException, ServletException
    {
    	HttpSession session = request.getSession(false);
    	
    	
    	if(session != null)
    	{
    		if(isDebugMode())
        	{
        		context.log("[CSRFGuard] doResponseEnforce: attempting to parse the HTML resource " + request.getRequestURI());
        	}
    		
        	String oToken = (String)session.getAttribute(tokenName);
        	
        	doToken(response, oToken);
    	}
    	else if(isDebugMode())
    	{
    		context.log("[CSRFGuard] doResponseEnforce: unable to verify the token as the session does not exist");
    	}
    }
    
	private void doToken(MutableHttpResponse response, String oToken)
	{
		try
		{
			Parser p = new Parser();
			
			p.setResource(new String(response.getContent()));
			
			NodeList nl = p.parse(null);
			NodeIterator itr = nl.elements();
			
			while(itr.hasMoreNodes())
			{
				Node n = (Node)itr.nextNode();
				processNodes(n, oToken);
			}
			
			response.setContent(nl.toHtml());
		}
		catch (Exception e)
		{
			context.log(e.getMessage(), e);
		}
	}
	
	private void processNodes(Node node, String oToken) throws ParserException
	{
		if(node instanceof LinkTag)
		{
			LinkTag link = (LinkTag)node;
			
			if(link.isHTTPLink() || link.isHTTPSLink())
			{
				String text = link.getAttribute("href");
				
				if(text.indexOf("?") != -1)
				{
					text = text.replace("?", "?" + tokenName + "=" + oToken + "&");
				}
				else
				{
					text = text + "?" + tokenName + "=" + oToken;
				}
				
				link.setLink(text);
			}
		}
		else if(node instanceof FormTag)
		{
			FormTag ft = (FormTag)node;
			InputTag it = new InputTag();
			
			it.setAttribute("type", "hidden");
			it.setAttribute("name", tokenName);
			it.setAttribute("value", oToken);
			
			NodeList children = ft.getChildren();
			
			children.add(it);
			
			ft.setChildren(children);
		}
		else if(node instanceof ScriptTag)
		{
			/**
			 * This is where we would handle JavaScript code
			 */
		}
		else if(node instanceof TagNode)
		{
	         TagNode tag = (TagNode)node;
	         NodeList nl = tag.getChildren();
	         
	         if(nl != null)
	         {
	        	 for(NodeIterator itr = nl.elements(); itr.hasMoreNodes(); processNodes(itr.nextNode(), oToken));
	         }
		}
	}
	
	private String getPRNGAlgorithm(FilterConfig config)
	{
		String s = config.getInitParameter(PARAM_PRNG_ALGORITHM);
		
		if(s == null)
		{
			s = CryptoUtil.DEFAULT_PRNG;
		}
		
		return s;
	}
	
	private int getTokenLength(FilterConfig config) throws NumberFormatException
	{
		String s = config.getInitParameter(PARAM_TOKEN_LENGTH);
		
		if(s != null)
		{
			tokenLength = Integer.parseInt(s);
		}
		else
		{
			tokenLength = 16;
		}
		
		return tokenLength;
	}
	
	private String getErrorPage(FilterConfig config)
	{
		return config.getInitParameter(PARAM_ERROR_PAGE);
	}
	
	private String getTokenName(FilterConfig config)
	{
		String s = config.getInitParameter(PARAM_TOKEN_NAME);
		
		return (s == null ? "__tokenName" : s);
	}
	
	private boolean getDebug(FilterConfig config)
	{
		String s = config.getInitParameter(PARAM_DEBUG);
		
		return (s == null ? false : Boolean.parseBoolean(s));
	}
	
	private boolean isDebugMode()
	{
		return debug;
	}
}
